"""Generate biased subsets."""
from typing import Tuple, Sequence, Dict

import pandas as pd
import numpy as np

from ethicml.preprocessing.domain_adaptation import query_dt, make_valid_variable_name
from ethicml.utility.data_structures import concat_dt, DataTuple
from .train_test_split import ProportionalSplit, DataSplitter

__all__ = [
    'BiasedDebiasedSubsets',
    'BiasedSubset',
    'get_biased_and_debiased_subsets',
    'get_biased_subset',
]


class BiasedSubset(DataSplitter):
    """Split the given data into a biased subset and a normal subset."""

    def __init__(
        self,
        unbiased_pcnt: float,
        mixing_factors: Sequence[float] = (0,),
        seed: int = 42,
        data_efficient: bool = True,
    ):
        """Init BiasedSplit.

        Args:
            mixing_factors: List of mixing factors; they are chosen based on the split ID
            unbiased_pcnt: how much of the data should be reserved for the unbiased subset
            seed: random seed for the splitting
            data_efficient: if True, try to keep as many data points as possible
        """
        super().__init__()
        self.unbiased_pcnt = unbiased_pcnt
        self.mixing_factors = mixing_factors
        self.seed = seed
        self.data_efficient = data_efficient

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do BiasedSplit."""
        mixing_factor = self.mixing_factors[split_id]
        biased, unbiased = get_biased_subset(
            data, mixing_factor, self.unbiased_pcnt, self.seed, self.data_efficient
        )
        return biased, unbiased, {'mix_fact': mixing_factor}


def get_biased_subset(
    data: DataTuple,
    mixing_factor: float,
    unbiased_pcnt: float,
    seed: int = 42,
    data_efficient: bool = True,
) -> Tuple[DataTuple, DataTuple]:
    """Split the given data into a biased subset and a normal subset.

    The two subsets don't generally sum up to the whole set.

    Example behavior:

    * mixing_factor=0.0: in biased, s=y everywhere; unbiased is just a subset of `data`
    * mixing_factor=0.5: biased and unbiased are both just subsets of `data`
    * mixing_factor=1.0: in biased, s!=y everywhere; unbiased is just a subset of `data`

    Args:
        data: data in form of a DataTuple
        mixing_factor: How much of the debiased data should be mixed into the biased subset? If this
                       factor is 0, the biased subset is maximally biased.
        unbiased_pcnt: how much of the data should be reserved for the unbiased subset
        seed: random seed for the splitting
        data_efficient: if True, try to keep as many data points as possible

    Returns:
        biased and unbiased dataset
    """
    assert 0 <= mixing_factor <= 1, f"mixing_factor: {mixing_factor}"
    assert 0 <= unbiased_pcnt <= 1, f"unbiased_pcnt: {unbiased_pcnt}"
    s_name = data.s.columns[0]
    y_name = data.y.columns[0]

    normal_subset, for_biased_subset = _random_split(data, first_pcnt=unbiased_pcnt, seed=seed)

    sy_equal, sy_opposite = _get_sy_equal_and_opp(for_biased_subset, s_name, y_name)

    mix_fact = mixing_factor  # how much of sy_opp should be mixed into the biased subset

    if data_efficient:
        if mix_fact < 0.5:
            sy_equal_fraction = 1.0
            sy_opp_fraction = 2 * mix_fact
        else:
            sy_equal_fraction = 2 * (1 - mix_fact)
            sy_opp_fraction = 1.0
    else:
        sy_equal_fraction = 1 - mix_fact
        sy_opp_fraction = mix_fact

    sy_equal_for_biased_ss, _ = _random_split(sy_equal, first_pcnt=sy_equal_fraction, seed=seed)
    sy_opp_for_biased_ss, _ = _random_split(sy_opposite, first_pcnt=sy_opp_fraction, seed=seed)

    biased_subset = concat_dt(
        [sy_equal_for_biased_ss, sy_opp_for_biased_ss], axis='index', ignore_index=True
    )

    if mix_fact == 0:
        # s and y should be very correlated in the biased subset
        assert all(biased_subset.s[s_name] == biased_subset.y[y_name])

    biased_subset = biased_subset.replace(name=f"{data.name} - Biased (tm={mixing_factor})")
    normal_subset = normal_subset.replace(name=f"{data.name} - Subset (tm={mixing_factor})")
    return biased_subset, normal_subset


class BiasedDebiasedSubsets(DataSplitter):
    """Split the given data into a biased subset and a debiased subset."""

    def __init__(
        self,
        unbiased_pcnt: float,
        mixing_factors: Sequence[float] = (0,),
        seed: int = 42,
        fixed_unbiased: bool = True,
    ):
        """Init BiasedDebiasedSplit.

        Args:
            mixing_factors: List of mixing factors; they are chosen based on the split ID
            unbiased_pcnt: how much of the data should be reserved for the unbiased subset
            seed: random seed for the splitting
            fixed_unbiased: if True, then the unbiased dataset is independent from the mixing factor
        """
        super().__init__()
        self.unbiased_pcnt = unbiased_pcnt
        self.mixing_factors = mixing_factors
        self.seed = seed
        self.fixed_unbiased = fixed_unbiased

    def __call__(
        self, data: DataTuple, split_id: int = 0
    ) -> Tuple[DataTuple, DataTuple, Dict[str, float]]:
        """Do Biased and Debiased Split."""
        mixing_factor = self.mixing_factors[split_id]
        biased, unbiased = get_biased_and_debiased_subsets(
            data, mixing_factor, self.unbiased_pcnt, self.seed, self.fixed_unbiased
        )
        return biased, unbiased, {'mix_fact': mixing_factor}


def get_biased_and_debiased_subsets(
    data: DataTuple,
    mixing_factor: float,
    unbiased_pcnt: float,
    seed: int = 42,
    fixed_unbiased: bool = True,
) -> Tuple[DataTuple, DataTuple]:
    """Split the given data into a biased subset and a debiased subset.

    In contrast to :func:`get_biased_subset()`, this function makes the unbiased subset *really*
    unbiased.

    The two subsets don't generally sum up to the whole set.

    Example behavior:

    * mixing_factor=0.0: in biased, s=y everywhere; in debiased, 50% s=y and 50% s!=y
    * mixing_factor=0.5: biased is just a subset of `data`; in debiased, 50% s=y and 50% s!=y
    * mixing_factor=1.0: in biased, s!=y everywhere; in debiased, 50% s=y and 50% s!=y

    Args:
        data: data in form of a DataTuple
        mixing_factor: How much of the debiased data should be mixed into the biased subset? If this
                       factor is 0, the biased subset is maximally biased.
        unbiased_pcnt: how much of the data should be reserved for the unbiased subset
        seed: random seed for the splitting
        fixed_unbiased: if True, then the unbiased dataset is independent from the mixing factor

    Returns:
        biased and unbiased dataset
    """
    assert 0 <= mixing_factor <= 1, f"mixing_factor: {mixing_factor}"
    assert 0 <= unbiased_pcnt <= 1, f"unbiased_pcnt: {unbiased_pcnt}"
    s_name = data.s.columns[0]
    y_name = data.y.columns[0]
    sy_equal, sy_opposite = _get_sy_equal_and_opp(data, s_name, y_name)

    # how much of sy_equal should be reserved for the biased subset:

    if fixed_unbiased:
        sy_equal_for_debiased_ss, sy_equal_remaining = _random_split(
            sy_equal, first_pcnt=unbiased_pcnt, seed=seed
        )
        sy_opp_for_debiased_ss, sy_opp_remaining = _random_split(
            sy_opposite, first_pcnt=unbiased_pcnt, seed=seed
        )
        if mixing_factor < 0.5:
            sy_equal_fraction = 1.0
            sy_opp_fraction = 2 * mixing_factor
        else:
            sy_equal_fraction = 2 * (1 - mixing_factor)
            sy_opp_fraction = 1.0

        sy_equal_for_biased_ss, _ = _random_split(
            sy_equal_remaining, first_pcnt=sy_equal_fraction, seed=seed
        )
        sy_opp_for_biased_ss, _ = _random_split(
            sy_opp_remaining, first_pcnt=sy_opp_fraction, seed=seed
        )
    else:
        biased_pcnt = 1 - unbiased_pcnt
        sy_equal_for_biased_ss, sy_equal_for_debiased_ss = _random_split(
            sy_equal, first_pcnt=biased_pcnt * (1 - mixing_factor), seed=seed
        )
        sy_opp_for_biased_ss, sy_opp_for_debiased_ss = _random_split(
            sy_opposite, first_pcnt=biased_pcnt * mixing_factor, seed=seed
        )

    biased_subset = concat_dt(
        [sy_equal_for_biased_ss, sy_opp_for_biased_ss], axis='index', ignore_index=True
    )

    # the debiased set is constructed from two sets of the same size
    min_size = min(len(sy_equal_for_debiased_ss), len(sy_opp_for_debiased_ss))

    def _get_equal_sized_subset(df: pd.DataFrame) -> pd.DataFrame:
        return df.sample(n=min_size, random_state=seed)

    debiased_subset_part1 = sy_equal_for_debiased_ss.apply_to_joined_df(_get_equal_sized_subset)
    debiased_subset_part2 = sy_opp_for_debiased_ss.apply_to_joined_df(_get_equal_sized_subset)
    debiased_subset = concat_dt(
        [debiased_subset_part1, debiased_subset_part2], axis='index', ignore_index=True
    )

    # s and y should not be correlated in the debiased subset
    assert abs(debiased_subset.s[s_name].corr(debiased_subset.y[y_name])) < 0.5

    if mixing_factor == 0:
        # s and y should be very correlated in the biased subset
        assert biased_subset.s[s_name].corr(biased_subset.y[y_name]) > 0.99

    biased_subset = biased_subset.replace(name=f"{data.name} - Biased (tm={mixing_factor})")
    debiased_subset = debiased_subset.replace(name=f"{data.name} - Debiased (tm={mixing_factor})")
    return biased_subset, debiased_subset


def _random_split(data: DataTuple, first_pcnt: float, seed: int) -> Tuple[DataTuple, DataTuple]:
    splitter = ProportionalSplit(train_percentage=first_pcnt, start_seed=seed)
    return splitter(data)[0:2]


def _get_sy_equal_and_opp(data: DataTuple, s_name: str, y_name: str) -> Tuple[DataTuple, DataTuple]:
    """Get the subset where s and y are equal and the subset where they are opposite."""
    s_name = make_valid_variable_name(s_name)
    y_name = make_valid_variable_name(y_name)
    s_values = np.unique(data.s.to_numpy())
    y_values = np.unique(data.y.to_numpy())
    assert len(s_values) == 2, "function only works with binary sensitive attribute"
    assert len(y_values) == 2, "function only works with binary labels"
    # we implicitly assume that the minimum value of s and y are equal; same for the maximum value
    s_0, s_1 = s_values
    y_0, y_1 = y_values
    # datapoints where s and y are the same
    s_and_y_equal = query_dt(
        data, f"({s_name} == {s_0} & {y_name} == {y_0}) | ({s_name} == {s_1} & {y_name} == {y_1})"
    )
    # datapoints where they are not the same
    s_and_y_opposite = query_dt(
        data, f"({s_name} == {s_0} & {y_name} == {y_1}) | ({s_name} == {s_1} & {y_name} == {y_0})"
    )
    return s_and_y_equal, s_and_y_opposite



"""Installable model.

This is a kind of complicated model, but it's incredibly useful.
Say you find a paper from a few years ago with code. It's not unreasonable that there might
be dependency clashes, python clashes, clashes galore. This approach downloads a model, runs it
in its own venv and makes everyone happy.
"""
import os
from pathlib import Path
from typing import List, Optional
import shutil
import subprocess

import git

from ethicml.algorithms.inprocess.in_algorithm import InAlgorithmAsync
from ethicml.utility.data_structures import PathTuple, TestPathTuple


class InstalledModel(InAlgorithmAsync):
    """the model that does the magic."""

    def __init__(
        self,
        dir_name: str,
        top_dir: str,
        url: Optional[str] = None,
        executable: Optional[str] = None,
    ):
        """Download code from given URL and create Pip environment with Pipfile found in the code.

        Args:
            dir_name: where to download the code to (can be chosen freely)
            top_dir: top directory of the repository where the Pipfile can be found (this is usually
                     simply the last part of the repository URL)
            url: (optional) URL of the repository
            executable: (optional) path to a Python executable
        """
        # QUESTION: do we really need `store_dir`? we could also just clone the code into "."
        self._store_dir: Path = Path(".") / dir_name  # directory where code and venv are stored
        self._top_dir: str = top_dir

        if url is not None:
            # download code
            self._clone_directory(url)

        if executable is None:
            # create virtual environment
            self._create_venv()
            self.__executable = str(self._code_path.resolve() / ".venv" / "bin" / "python")
        else:
            self.__executable = executable
        super().__init__()

    @property
    def _code_path(self) -> Path:
        """Path to where the code of the model is located."""
        return self._store_dir / self._top_dir

    @property
    def _executable(self) -> str:
        return self.__executable

    def _clone_directory(self, url: str) -> None:
        """Clones the repo from `url` into `self._store_dir`."""
        if not self._store_dir.exists():
            self._store_dir.mkdir()
            git.Git(self._store_dir).clone(url)

    def _create_venv(self) -> None:
        """Creates a venv based on the Pipfile in the repository."""
        venv_directory = self._code_path / ".venv"
        if not venv_directory.exists():
            environ = os.environ.copy()
            environ["PIPENV_IGNORE_VIRTUALENVS"] = "1"
            environ["PIPENV_VENV_IN_PROJECT"] = "true"
            environ["PIPENV_YES"] = "true"
            environ["PIPENV_PIPFILE"] = str(self._code_path / "Pipfile")

            subprocess.check_call("pipenv install", env=environ, shell=True)

    def remove(self) -> None:
        """Removes the directory that we created in _clone_directory()."""
        try:
            shutil.rmtree(self._store_dir)
        except OSError as excep:
            print("Error: %s - %s." % (excep.filename, excep.strerror))

    def _script_command(
        self, train_paths: PathTuple, test_paths: TestPathTuple, pred_path: Path
    ) -> (List[str]):
        return []  # pylint was complaining when I didn't return anything here...

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        raise NotImplementedError



"""Implementation of logistic regression (actually just a wrapper around sklearn)"""
import sys
from pathlib import Path

from sklearn.linear_model import LogisticRegression
import pandas as pd


def main():
    """This function runs the logistic regression model as a standalone program"""
    args = sys.argv[1:]
    x, y = pd.read_feather(Path(args[0])), pd.read_feather(Path(args[1]))
    test_x = pd.read_feather(Path(args[2]))
    clf = LogisticRegression(solver="liblinear", random_state=888, multi_class='auto')
    clf.fit(x, y.to_numpy().ravel())
    predictions = pd.DataFrame(clf.predict(test_x), columns=["preds"])
    predictions.to_feather(Path(args[3]))


if __name__ == "__main__":
    main()



"""Implementation of logistic regression (actually just a wrapper around sklearn)."""
import random

import numpy as np
import pandas as pd

from fairlearn.reductions import (
    ExponentiatedGradient,
    DemographicParity,
    EqualizedOdds,
    ConditionalSelectionRate,
)
from sklearn.linear_model import LogisticRegression

from ethicml.utility.data_structures import DataTuple, TestTuple, FairnessType, ClassifierType
from ethicml.algorithms.inprocess.svm import select_svm
from .utils import InAlgoArgs, load_data_from_flags, save_predictions


class AgarwalArgs(InAlgoArgs):
    """Args for the Agarwal implementation."""

    classifier: ClassifierType
    fairness: FairnessType
    eps: float
    iters: int
    C: float
    kernel: str


def train_and_predict(train: DataTuple, test: TestTuple, args: AgarwalArgs):
    """Train a logistic regression model and compute predictions on the given test data."""
    random.seed(888)
    np.random.seed(888)

    fairness_class: ConditionalSelectionRate
    if args.fairness == "DP":
        fairness_class = DemographicParity()
    else:
        fairness_class = EqualizedOdds()

    if args.classifier == "SVM":
        model = select_svm(args.C, args.kernel)
    else:
        model = LogisticRegression(solver="liblinear", random_state=888, max_iter=5000, C=args.C)

    data_x = train.x
    data_y = train.y[train.y.columns[0]]
    data_a = train.s[train.s.columns[0]]

    exponentiated_gradient = ExponentiatedGradient(
        model, constraints=fairness_class, eps=args.eps, T=args.iters
    )
    exponentiated_gradient.fit(data_x, data_y, sensitive_features=data_a)

    randomized_predictions = exponentiated_gradient.predict(test.x)
    preds = pd.DataFrame(randomized_predictions, columns=["preds"])

    min_class_label = train.y[train.y.columns[0]].min()
    if preds["preds"].min() != preds["preds"].max():
        preds = preds.replace(preds["preds"].min(), min_class_label)
    return preds


def main():
    """This function runs the Agarwal model as a standalone program."""
    args: AgarwalArgs = AgarwalArgs().parse_args()
    train, test = load_data_from_flags(args)
    save_predictions(train_and_predict(train, test, args), args)


if __name__ == "__main__":
    main()



"""Transformations that act differentlly depending on the label."""

from typing import List

import numpy as np
import torch
from torch import Tensor
from torch import jit

__all__ = ["LdAugmentation", "LdColorizer"]


class LdAugmentation(jit.ScriptModule):
    """Base class for label-dependent augmentations."""

    @jit.script_method
    def _augment(self, data: Tensor, labels: Tensor) -> Tensor:
        """Augment the input data in a label-dependent fashion.

        Args:
            data: Tensor. Input data to be augmented.
            labels: Tensor. Labels on which the augmentations are conditioned.

        Returns: Tensor. Augmented data.
        """
        return data

    def __call__(self, data: Tensor, labels: Tensor) -> Tensor:
        """Apply the augment method to the input data.

        Args:
            data: Tensor. Input data to be augmented.
            labels: Tensor. Labels on which the augmentations are conditioned.

        Returns: Tensor. Augmented data.
        """
        return self._augment(data, labels)


class LdColorizer(LdAugmentation):
    """Transform that colorizes images."""

    __constants__ = ["min_val", "max_val", "binarize", "black", "background", "seed"]

    def __init__(
        self,
        min_val: float = 0.0,
        max_val: float = 1.0,
        scale: float = 0.02,
        binarize: bool = False,
        background: bool = False,
        black: bool = True,
        seed: int = 42,
        greyscale: bool = False,
    ):
        """Colorizes a grayscale image by sampling colors from multivariate normal distributions.

        The distribution is centered on predefined means and standard deviation determined by the
        scale argument.

        Args:
            min_val: Minimum value the input data can take (needed for clamping). Defaults to 0.
            max_val: Maximum value the input data can take (needed for clamping). Defaults to 1.
            scale: Standard deviation of the multivariate normal distributions from which
                the colors are drawn. Lower values correspond to higher bias. Defaults to 0.02.
            binarize: Whether the binarize the grayscale data before colorisation. Defaults to False
            background: Whether to color the background instead of the foreground. Defaults to False
            black: Whether not to invert the black. Defaults to True.
            seed: Random seed used for sampling colors. Defaults to 42.
            greyscale: Whether to greyscale the colorised images. Defaults to False.
        """
        super(LdColorizer, self).__init__()
        self.min_val = min_val
        self.max_val = max_val
        self.scale = scale * np.eye(3)
        self.binarize = binarize
        self.background = background
        self.black = black
        self.greyscale = greyscale

        # create a local random state that won't affect the global random state of the training
        self.random_state = np.random.RandomState(seed)

        colors = [
            (0, 255, 255),
            (0, 0, 255),  # blue
            (255, 0, 255),
            (0, 128, 0),
            (0, 255, 0),  # green
            (128, 0, 0),
            (0, 0, 128),
            (128, 0, 128),
            (255, 0, 0),  # red
            (255, 255, 0),
        ]  # yellow

        self.palette = [np.divide(color, 255) for color in colors]

    def _sample_color(self, mean_color_values: np.ndarray) -> np.ndarray:
        return np.clip(self.random_state.multivariate_normal(mean_color_values, self.scale), 0, 1)

    def _augment(self, data: Tensor, labels: Tensor) -> Tensor:
        """Apply the transformation.

        Args:
            data (Tensor): (Grayscale) data samples to be colorized.
            labels (Tensor): Index indicating the colour distribution
            from which to sample for each data sample.

        Returns:
            Tensor: Colorized tensor.
        """
        labels = labels.numpy()

        mean_values = []
        colors_per_sample: List[np.ndarray] = []
        for label in labels:
            mean_value = self.palette[label]
            mean_values.append(mean_value)
            colors_per_sample.append(self._sample_color(mean_value))

        if self.binarize:
            data = (data > 0.5).float()

        color_tensor = Tensor(colors_per_sample).unsqueeze(-1).unsqueeze(-1)
        if self.background:
            if self.black:
                # colorful background, black digits
                augmented_data = (1 - data) * color_tensor
            else:
                # colorful background, white digits
                augmented_data = torch.clamp(data + color_tensor, 0, 1)
        else:
            if self.black:
                # black background, colorful digits
                augmented_data = data * color_tensor
            else:
                # white background, colorful digits
                augmented_data = 1 - data * (1 - color_tensor)

        if self.greyscale:
            augmented_data = augmented_data.mean(dim=1, keepdim=True)

        return augmented_data



"""Test that we can get some metrics on predictions"""

from typing import Tuple
import pandas as pd
import pytest
from pytest import approx

from ethicml.algorithms.inprocess import InAlgorithm, LRProb, SVM, LR, Kamiran
from ethicml.utility import DataTuple
from ethicml.data import Adult, NonBinaryToy, load_data
from ethicml.evaluators import (
    diff_per_sensitive_attribute,
    metric_per_sensitive_attribute,
    ratio_per_sensitive_attribute,
    run_metrics,
    MetricNotApplicable,
)
from ethicml.metrics import (
    Accuracy,
    AS,
    BCR,
    CV,
    F1,
    Metric,
    NMI,
    PPV,
    NPV,
    ProbNeg,
    ProbOutcome,
    ProbPos,
    TNR,
    TPR,
    Theil,
    Hsic,
    LabelOutOfBounds,
)
from ethicml.preprocessing import train_test_split
from tests.run_algorithm_test import get_train_test

RTOL = 1e-5  # relative tolerance when comparing two floats


@pytest.mark.parametrize(
    "name,metric,expected_value", [("Accuracy", Accuracy(), 0.89), ("F1", F1(), 0.893)]
)
def test_get_score_of_predictions(name: str, metric: Metric, expected_value: float) -> None:
    """test get acc of predictions"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    assert metric.name == name
    score = metric.score(predictions, test)
    assert score == approx(expected_value, abs=0.001)


def test_mni_preds_and_s():
    """test mni preds and s"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc: Metric = NMI(base='s')
    assert acc.name == "NMI preds and s"
    score = acc.score(predictions, test)
    assert score == pytest.approx(0.083, abs=0.001)


def test_accuracy_per_sens_attr():
    """test accuracy per sens attr"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, Accuracy())
    assert acc_per_sens == {'s_0': 0.905, 's_1': 0.875}


def test_anti_spurious() -> None:
    """test anti-spurious"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    metric: Metric = AS()
    assert metric.name == "anti_spurious"
    score = metric.score(predictions, test)
    assert score == pytest.approx(0.892, abs=0.001)


def test_probpos_per_sens_attr():
    """test probpos per sens attr"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, ProbPos())
    assert acc_per_sens == {'s_0': 0.335, 's_1': 0.67}


def test_proboutcome_per_sens_attr():
    """test proboutcome per sens attr"""
    train, test = get_train_test()
    model: InAlgorithm = LRProb()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, ProbOutcome())
    assert acc_per_sens == {
        's_0': pytest.approx(0.372, abs=0.001),
        's_1': pytest.approx(0.661, abs=0.001),
    }


def test_probneg_per_sens_attr():
    """test probneg per sens attr"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, ProbNeg())
    assert acc_per_sens == {'s_0': 0.665, 's_1': 0.33}


def test_acc_per_nonbinary_sens():
    """test acc per nonbinary sens"""
    data: DataTuple = load_data(Adult("Nationality"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, Accuracy())

    test_dict = {
        'native-country_Cambodia_0': pytest.approx(0.778, abs=0.001),
        'native-country_Cambodia_1': pytest.approx(0.500, abs=0.001),
        'native-country_Canada_0': pytest.approx(0.779, abs=0.001),
        'native-country_Canada_1': pytest.approx(0.595, abs=0.001),
        'native-country_China_0': pytest.approx(0.778, abs=0.001),
        'native-country_China_1': pytest.approx(0.800, abs=0.001),
        'native-country_Columbia_0': pytest.approx(0.778, abs=0.001),
        'native-country_Columbia_1': pytest.approx(1.000, abs=0.001),
        'native-country_Cuba_0': pytest.approx(0.778, abs=0.001),
        'native-country_Cuba_1': pytest.approx(0.826, abs=0.001),
        'native-country_Dominican-Republic_0': pytest.approx(0.778, abs=0.001),
        'native-country_Dominican-Republic_1': pytest.approx(0.933, abs=0.001),
        'native-country_Ecuador_0': pytest.approx(0.778, abs=0.001),
        'native-country_Ecuador_1': pytest.approx(0.800, abs=0.001),
        'native-country_El-Salvador_0': pytest.approx(0.778, abs=0.001),
        'native-country_El-Salvador_1': pytest.approx(1.000, abs=0.001),
        'native-country_England_0': pytest.approx(0.779, abs=0.001),
        'native-country_England_1': pytest.approx(0.600, abs=0.001),
        'native-country_France_0': pytest.approx(0.778, abs=0.001),
        'native-country_France_1': pytest.approx(0.500, abs=0.001),
        'native-country_Germany_0': pytest.approx(0.778, abs=0.001),
        'native-country_Germany_1': pytest.approx(0.679, abs=0.001),
        'native-country_Greece_0': pytest.approx(0.778, abs=0.001),
        'native-country_Greece_1': pytest.approx(0.636, abs=0.001),
        'native-country_Guatemala_0': pytest.approx(0.778, abs=0.001),
        'native-country_Guatemala_1': pytest.approx(0.952, abs=0.001),
        'native-country_Haiti_0': pytest.approx(0.778, abs=0.001),
        'native-country_Haiti_1': pytest.approx(0.929, abs=0.001),
        'native-country_Holand-Netherlands_0': pytest.approx(0.778, abs=0.001),
        'native-country_Honduras_0': pytest.approx(0.778, abs=0.001),
        'native-country_Honduras_1': pytest.approx(1.000, abs=0.001),
        'native-country_Hong_0': pytest.approx(0.778, abs=0.001),
        'native-country_Hong_1': pytest.approx(0.667, abs=0.001),
        'native-country_Hungary_0': pytest.approx(0.778, abs=0.001),
        'native-country_Hungary_1': pytest.approx(1.000, abs=0.001),
        'native-country_India_0': pytest.approx(0.778, abs=0.001),
        'native-country_India_1': pytest.approx(0.589, abs=0.001),
        'native-country_Iran_0': pytest.approx(0.778, abs=0.001),
        'native-country_Iran_1': pytest.approx(0.364, abs=0.001),
        'native-country_Ireland_0': pytest.approx(0.778, abs=0.001),
        'native-country_Ireland_1': pytest.approx(0.818, abs=0.001),
        'native-country_Italy_0': pytest.approx(0.778, abs=0.001),
        'native-country_Italy_1': pytest.approx(0.565, abs=0.001),
        'native-country_Jamaica_0': pytest.approx(0.778, abs=0.001),
        'native-country_Jamaica_1': pytest.approx(0.800, abs=0.001),
        'native-country_Japan_0': pytest.approx(0.778, abs=0.001),
        'native-country_Japan_1': pytest.approx(0.760, abs=0.001),
        'native-country_Laos_0': pytest.approx(0.778, abs=0.001),
        'native-country_Laos_1': pytest.approx(1.000, abs=0.001),
        'native-country_Mexico_0': pytest.approx(0.774, abs=0.001),
        'native-country_Mexico_1': pytest.approx(0.963, abs=0.001),
        'native-country_Nicaragua_0': pytest.approx(0.778, abs=0.001),
        'native-country_Nicaragua_1': pytest.approx(0.875, abs=0.001),
        'native-country_Outlying-US(Guam-USVI-etc)_0': pytest.approx(0.778, abs=0.001),
        'native-country_Outlying-US(Guam-USVI-etc)_1': pytest.approx(0.800, abs=0.001),
        'native-country_Peru_0': pytest.approx(0.778, abs=0.001),
        'native-country_Peru_1': pytest.approx(0.923, abs=0.001),
        'native-country_Philippines_0': pytest.approx(0.778, abs=0.001),
        'native-country_Philippines_1': pytest.approx(0.717, abs=0.001),
        'native-country_Poland_0': pytest.approx(0.778, abs=0.001),
        'native-country_Poland_1': pytest.approx(0.813, abs=0.001),
        'native-country_Portugal_0': pytest.approx(0.778, abs=0.001),
        'native-country_Portugal_1': pytest.approx(0.800, abs=0.001),
        'native-country_Puerto-Rico_0': pytest.approx(0.778, abs=0.001),
        'native-country_Puerto-Rico_1': pytest.approx(0.921, abs=0.001),
        'native-country_Scotland_0': pytest.approx(0.778, abs=0.001),
        'native-country_Scotland_1': pytest.approx(1.000, abs=0.001),
        'native-country_South_0': pytest.approx(0.778, abs=0.001),
        'native-country_South_1': pytest.approx(0.714, abs=0.001),
        'native-country_Taiwan_0': pytest.approx(0.778, abs=0.001),
        'native-country_Taiwan_1': pytest.approx(0.429, abs=0.001),
        'native-country_Thailand_0': pytest.approx(0.778, abs=0.001),
        'native-country_Thailand_1': pytest.approx(0.750, abs=0.001),
        'native-country_Trinadad&Tobago_0': pytest.approx(0.778, abs=0.001),
        'native-country_Trinadad&Tobago_1': pytest.approx(1.000, abs=0.001),
        'native-country_United-States_0': pytest.approx(0.806, abs=0.001),
        'native-country_United-States_1': pytest.approx(0.776, abs=0.001),
        'native-country_Vietnam_0': pytest.approx(0.778, abs=0.001),
        'native-country_Vietnam_1': pytest.approx(0.929, abs=0.001),
        'native-country_Yugoslavia_0': pytest.approx(0.778, abs=0.001),
        'native-country_Yugoslavia_1': pytest.approx(0.714, abs=0.001),
    }

    for key in acc_per_sens:
        assert acc_per_sens[key] == test_dict[key]


def test_acc_per_race():
    """test acc per race"""
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    acc_per_sens = metric_per_sensitive_attribute(predictions, test, Accuracy())
    test_dict = {
        'race_Amer-Indian-Eskimo_0': pytest.approx(0.78, abs=0.01),
        'race_Amer-Indian-Eskimo_1': pytest.approx(0.94, abs=0.01),
        'race_Asian-Pac-Islander_0': pytest.approx(0.78, abs=0.01),
        'race_Asian-Pac-Islander_1': pytest.approx(0.73, abs=0.01),
        'race_Black_0': pytest.approx(0.77, abs=0.01),
        'race_Black_1': pytest.approx(0.87, abs=0.01),
        'race_Other_0': pytest.approx(0.78, abs=0.01),
        'race_Other_1': pytest.approx(0.88, abs=0.01),
        'race_White_0': pytest.approx(0.85, abs=0.01),
        'race_White_1': pytest.approx(0.77, abs=0.01),
    }

    for key in acc_per_sens:
        assert acc_per_sens[key] == test_dict[key]


def test_tpr_diff():
    """test tpr diff"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert TPR().name == "TPR"
    assert tprs == {'s_0': pytest.approx(0.84, abs=0.01), 's_1': pytest.approx(0.88, abs=0.01)}
    tpr_diff = diff_per_sensitive_attribute(tprs)
    print(tpr_diff)
    assert tpr_diff["s_0-s_1"] == pytest.approx(0.04, abs=0.01)


def test_get_nmi_of_predictions():
    """test get nmi of predictions"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    nmi: Metric = NMI()
    assert nmi.name == "NMI preds and y"
    score = nmi.score(predictions, test)
    assert score == pytest.approx(0.50, abs=0.01)


def test_nmi_diff():
    """test nmi diff"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    nmis = metric_per_sensitive_attribute(predictions, test, NMI())
    assert NMI().name == "NMI preds and y"
    assert nmis == {'s_0': pytest.approx(0.52, abs=0.01), 's_1': pytest.approx(0.42, abs=0.01)}
    nmi_diff = diff_per_sensitive_attribute(nmis)
    assert nmi_diff["s_0-s_1"] == pytest.approx(0.10, abs=0.01)


def test_ppv_diff():
    """test ppv diff"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = metric_per_sensitive_attribute(predictions, test, PPV())
    assert PPV().name == "PPV"
    assert results == {'s_0': pytest.approx(0.88, abs=0.01), 's_1': pytest.approx(0.93, abs=0.01)}
    diff = diff_per_sensitive_attribute(results)
    assert diff["s_0-s_1"] == pytest.approx(0.05, abs=0.1)


def test_npv_diff():
    """test npv diff"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = metric_per_sensitive_attribute(predictions, test, NPV())
    assert NPV().name == "NPV"
    assert results == {'s_0': pytest.approx(0.91, abs=0.01), 's_1': pytest.approx(0.75, abs=0.01)}
    diff = diff_per_sensitive_attribute(results)
    assert diff["s_0-s_1"] == pytest.approx(0.15, abs=0.01)


def test_bcr_diff():
    """test bcr diff"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = metric_per_sensitive_attribute(predictions, test, BCR())
    assert BCR().name == "BCR"
    assert results == {'s_0': pytest.approx(0.89, abs=0.01), 's_1': pytest.approx(0.86, abs=0.01)}
    diff = diff_per_sensitive_attribute(results)
    assert diff["s_0-s_1"] == pytest.approx(0.02, abs=0.01)


def test_cv():
    """test cv"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    cross_val = CV()
    score = cross_val.score(predictions, test)
    assert CV().name == "CV"
    assert score == 0.665


def test_theil():
    """test theil"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    theil = Theil()
    score = theil.score(predictions, test)
    assert Theil().name == "Theil_Index"
    assert score == pytest.approx(0.08, abs=0.01)

    model = LR()
    predictions = model.run(train, test)
    theil = Theil()
    score = theil.score(predictions, test)
    assert score == pytest.approx(0.07, abs=0.01)

    model = Kamiran()
    predictions = model.run(train, test)
    theil = Theil()
    score = theil.score(predictions, test)
    assert score == pytest.approx(0.08, abs=0.01)


def test_theil_per_sens_attr():
    """test theil per sens attr"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    theil_per_sens = metric_per_sensitive_attribute(predictions, test, Theil())
    assert theil_per_sens == {
        's_0': pytest.approx(0.07, abs=0.01),
        's_1': pytest.approx(0.10, abs=0.01),
    }


def test_hsic():
    """test hsic"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    hsic = Hsic()
    score = hsic.score(predictions, test)
    assert Hsic().name == "HSIC"
    assert score == pytest.approx(0.02, abs=0.01)


def test_use_appropriate_metric():
    """test use appropriate metric"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    with pytest.raises(MetricNotApplicable):
        metric_per_sensitive_attribute(predictions, test, CV())


def test_tnr_diff():
    """test tnr diff"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    nmis = metric_per_sensitive_attribute(predictions, test, TNR())
    assert NMI().name == "NMI preds and y"
    assert nmis == {'s_0': pytest.approx(0.93, abs=0.01), 's_1': pytest.approx(0.84, abs=0.01)}
    nmi_diff = diff_per_sensitive_attribute(nmis)
    assert nmi_diff["s_0-s_1"] == pytest.approx(0.09, abs=0.01)


def test_run_metrics():
    """test run metrics"""
    train, test = get_train_test()
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run(train, test)
    results = run_metrics(predictions, test, [CV()], [TPR()])
    assert len(results) == 5
    assert results['TPR_s_0'] == approx(0.842857, RTOL)
    assert results['TPR_s_1'] == approx(0.886525, RTOL)
    assert results['TPR_s_0-s_1'] == approx(abs(0.842857 - 0.886525), RTOL)
    assert results['TPR_s_0/s_1'] == approx(0.842857 / 0.886525, RTOL)
    assert results['CV'] == approx(0.665)


def test_nmi_diff_non_binary_race():
    """test nmi diff non binary race"""
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    nmis = metric_per_sensitive_attribute(predictions, test, NMI())
    assert NMI().name == "NMI preds and y"
    test_dict = {
        'race_Amer-Indian-Eskimo_0': pytest.approx(0.14, abs=0.01),
        'race_Amer-Indian-Eskimo_1': pytest.approx(0.41, abs=0.01),
        'race_Asian-Pac-Islander_0': pytest.approx(0.14, abs=0.01),
        'race_Asian-Pac-Islander_1': pytest.approx(0.10, abs=0.01),
        'race_Black_0': pytest.approx(0.14, abs=0.01),
        'race_Black_1': pytest.approx(0.15, abs=0.01),
        'race_Other_0': pytest.approx(0.14, abs=0.01),
        'race_Other_1': pytest.approx(0.18, abs=0.01),
        'race_White_0': pytest.approx(0.16, abs=0.01),
        'race_White_1': pytest.approx(0.14, abs=0.01),
    }

    for key in nmis:
        assert nmis[key] == test_dict[key]

    nmis_to_check = {
        k: nmis[k]
        for k in (
            'race_Amer-Indian-Eskimo_1',
            'race_Asian-Pac-Islander_1',
            'race_Black_1',
            'race_Other_1',
            'race_White_1',
        )
    }
    nmi_diff = diff_per_sensitive_attribute(nmis_to_check)
    test_dict = {
        'race_Amer-Indian-Eskimo_1-race_Asian-Pac-Islander_1': pytest.approx(0.31, abs=0.01),
        'race_Amer-Indian-Eskimo_1-race_Black_1': pytest.approx(0.26, abs=0.01),
        'race_Amer-Indian-Eskimo_1-race_Other_1': pytest.approx(0.23, abs=0.01),
        'race_Amer-Indian-Eskimo_1-race_White_1': pytest.approx(0.27, abs=0.01),
        'race_Asian-Pac-Islander_1-race_Black_1': pytest.approx(0.04, abs=0.01),
        'race_Asian-Pac-Islander_1-race_Other_1': pytest.approx(0.08, abs=0.01),
        'race_Asian-Pac-Islander_1-race_White_1': pytest.approx(0.03, abs=0.01),
        'race_Black_1-race_Other_1': pytest.approx(0.03, abs=0.01),
        'race_Black_1-race_White_1': pytest.approx(0.02, abs=0.01),
        'race_Other_1-race_White_1': pytest.approx(0.04, abs=0.01),
    }

    for key in nmi_diff:
        assert nmi_diff[key] == test_dict[key]


def test_tpr_diff_non_binary_race():
    """test tpr diff non binary race"""
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert TPR().name == "TPR"
    test_dict = {
        'race_Amer-Indian-Eskimo_0': pytest.approx(0.16, abs=0.01),
        'race_Amer-Indian-Eskimo_1': pytest.approx(0.37, abs=0.01),
        'race_Asian-Pac-Islander_0': pytest.approx(0.16, abs=0.01),
        'race_Asian-Pac-Islander_1': pytest.approx(0.12, abs=0.01),
        'race_Black_0': pytest.approx(0.16, abs=0.01),
        'race_Black_1': pytest.approx(0.13, abs=0.01),
        'race_Other_0': pytest.approx(0.16, abs=0.01),
        'race_Other_1': pytest.approx(0.12, abs=0.01),
        'race_White_0': pytest.approx(0.14, abs=0.01),
        'race_White_1': pytest.approx(0.16, abs=0.01),
    }

    for key in tprs:
        assert tprs[key] == test_dict[key]

    tprs_to_check = {
        k: tprs[k]
        for k in (
            'race_Amer-Indian-Eskimo_1',
            'race_Asian-Pac-Islander_1',
            'race_Black_1',
            'race_Other_1',
            'race_White_1',
        )
    }
    tpr_diff = diff_per_sensitive_attribute(tprs_to_check)
    test_dict = {
        'race_Amer-Indian-Eskimo_1-race_Asian-Pac-Islander_1': pytest.approx(0.25, abs=0.01),
        'race_Amer-Indian-Eskimo_1-race_Black_1': pytest.approx(0.23, abs=0.01),
        'race_Amer-Indian-Eskimo_1-race_Other_1': pytest.approx(0.25, abs=0.01),
        'race_Amer-Indian-Eskimo_1-race_White_1': pytest.approx(0.20, abs=0.01),
        'race_Asian-Pac-Islander_1-race_Black_1': pytest.approx(0.01, abs=0.01),
        'race_Asian-Pac-Islander_1-race_Other_1': pytest.approx(0.00, abs=0.01),
        'race_Asian-Pac-Islander_1-race_White_1': pytest.approx(0.04, abs=0.01),
        'race_Black_1-race_Other_1': pytest.approx(0.01, abs=0.01),
        'race_Black_1-race_White_1': pytest.approx(0.04, abs=0.01),
        'race_Other_1-race_White_1': pytest.approx(0.04, abs=0.01),
    }

    for key in tpr_diff:
        assert tpr_diff[key] == test_dict[key]


def test_tpr_ratio_non_binary_race():
    """test tpr ratio non binary race"""
    data: DataTuple = load_data(Adult("Race"))
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert TPR().name == "TPR"
    test_dict = {
        'race_Amer-Indian-Eskimo_0': pytest.approx(0.16, abs=0.01),
        'race_Amer-Indian-Eskimo_1': pytest.approx(0.37, abs=0.01),
        'race_Asian-Pac-Islander_0': pytest.approx(0.16, abs=0.01),
        'race_Asian-Pac-Islander_1': pytest.approx(0.12, abs=0.01),
        'race_Black_0': pytest.approx(0.16, abs=0.01),
        'race_Black_1': pytest.approx(0.13, abs=0.01),
        'race_Other_0': pytest.approx(0.16, abs=0.01),
        'race_Other_1': pytest.approx(0.12, abs=0.01),
        'race_White_0': pytest.approx(0.14, abs=0.01),
        'race_White_1': pytest.approx(0.16, abs=0.01),
    }

    for key in tprs:
        assert tprs[key] == test_dict[key]

    tprs_to_check = {
        k: tprs[k]
        for k in (
            'race_Amer-Indian-Eskimo_1',
            'race_Asian-Pac-Islander_1',
            'race_Black_1',
            'race_Other_1',
            'race_White_1',
        )
    }
    tpr_diff = ratio_per_sensitive_attribute(tprs_to_check)
    test_dict = {
        'race_Amer-Indian-Eskimo_1/race_Asian-Pac-Islander_1': pytest.approx(0.32, abs=0.1),
        'race_Amer-Indian-Eskimo_1/race_Black_1': pytest.approx(0.37, abs=0.1),
        'race_Amer-Indian-Eskimo_1/race_Other_1': pytest.approx(0.33, abs=0.1),
        'race_Amer-Indian-Eskimo_1/race_White_1': pytest.approx(0.44, abs=0.1),
        'race_Asian-Pac-Islander_1/race_Black_1': pytest.approx(0.88, abs=0.1),
        'race_Asian-Pac-Islander_1/race_Other_1': pytest.approx(0.97, abs=0.1),
        'race_Asian-Pac-Islander_1/race_White_1': pytest.approx(0.72, abs=0.1),
        'race_Black_1/race_Other_1': pytest.approx(0.91, abs=0.1),
        'race_Black_1/race_White_1': pytest.approx(0.74, abs=0.1),
        'race_Other_1/race_White_1': pytest.approx(0.74, abs=0.1),
    }

    for key in tpr_diff:
        assert tpr_diff[key] == test_dict[key]


def test_nb_acc():
    """test nb acc"""
    data: DataTuple = load_data(NonBinaryToy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    acc_score = Accuracy().score(predictions, test)
    assert acc_score == 0.1
    accs = metric_per_sensitive_attribute(predictions, test, Accuracy())
    assert accs == {'sens_0': pytest.approx(0.09, abs=0.1), 'sens_1': pytest.approx(0.11, abs=0.1)}
    model = LR()
    predictions = model.run_test(train, test)
    acc_score = Accuracy().score(predictions, test)
    assert acc_score == 0.7
    accs = metric_per_sensitive_attribute(predictions, test, Accuracy())
    assert accs == {'sens_0': pytest.approx(0.72, abs=0.1), 'sens_1': pytest.approx(0.66, abs=0.1)}


def test_nb_tpr():
    """test nb tpr"""
    data: DataTuple = load_data(NonBinaryToy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tpr_score = TPR(pos_class=1).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=2).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=3).score(predictions, test)
    assert tpr_score == 1.0
    tpr_score = TPR(pos_class=4).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=5).score(predictions, test)
    assert tpr_score == 0.0

    with pytest.raises(LabelOutOfBounds):
        tpr_score = TPR(pos_class=0).score(predictions, test)

    accs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert accs == {'sens_0': pytest.approx(0.0, abs=0.1), 'sens_1': pytest.approx(0.0, abs=0.1)}

    model = LR()
    predictions = model.run_test(train, test)

    print(list([(k, z) for k, z in zip(predictions.values, test.y.values) if k != z]))

    tpr_score = TPR(pos_class=1).score(predictions, test)
    assert tpr_score == 1.0
    tpr_score = TPR(pos_class=2).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=3).score(predictions, test)
    assert tpr_score == 0.0
    tpr_score = TPR(pos_class=4).score(predictions, test)
    assert tpr_score == 1.0
    tpr_score = TPR(pos_class=5).score(predictions, test)
    assert tpr_score == 1.0

    with pytest.raises(LabelOutOfBounds):
        tpr_score = TPR(pos_class=0).score(predictions, test)

    tprs = metric_per_sensitive_attribute(predictions, test, TPR())
    assert tprs == {'sens_0': pytest.approx(1.0, abs=0.1), 'sens_1': pytest.approx(1.0, abs=0.1)}


def test_nb_tnr():
    """test nb tnr"""
    data: DataTuple = load_data(NonBinaryToy())
    train_test: Tuple[DataTuple, DataTuple] = train_test_split(data)
    train, test = train_test
    model: InAlgorithm = SVM()
    predictions: pd.DataFrame = model.run_test(train, test)
    tnr_score = TNR(pos_class=1).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=2).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=3).score(predictions, test)
    assert tnr_score == 0.0
    tnr_score = TNR(pos_class=4).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=5).score(predictions, test)
    assert tnr_score == 1.0

    with pytest.raises(LabelOutOfBounds):
        tnr_score = TNR(pos_class=0).score(predictions, test)

    accs = metric_per_sensitive_attribute(predictions, test, TNR())
    assert accs == {'sens_0': pytest.approx(1.0, abs=0.1), 'sens_1': pytest.approx(1.0, abs=0.1)}

    model = LR()
    predictions = model.run_test(train, test)

    print(list([(k, z) for k, z in zip(predictions.values, test.y.values) if k != z]))

    tnr_score = TNR(pos_class=1).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=2).score(predictions, test)
    assert tnr_score == 1.0
    tnr_score = TNR(pos_class=3).score(predictions, test)
    assert tnr_score == pytest.approx(0.7, abs=0.1)
    tnr_score = TNR(pos_class=4).score(predictions, test)
    assert tnr_score == pytest.approx(0.85, abs=0.1)
    tnr_score = TNR(pos_class=5).score(predictions, test)
    assert tnr_score == 1.0

    with pytest.raises(LabelOutOfBounds):
        tnr_score = TNR(pos_class=0).score(predictions, test)

    tnrs = metric_per_sensitive_attribute(predictions, test, TNR())
    assert tnrs == {'sens_0': pytest.approx(1.0, abs=0.1), 'sens_1': pytest.approx(1.0, abs=0.1)}



"""Post-processing method by Hardt et al."""
import numpy as np
from numpy.random import RandomState
import pandas as pd
from scipy.optimize import linprog, OptimizeResult

from ethicml.common import implements
from ethicml.utility.data_structures import DataTuple, TestTuple
from ethicml.metrics import TPR, TNR
from ethicml.evaluators.per_sensitive_attribute import metric_per_sensitive_attribute
from .post_algorithm import PostAlgorithm


class Hardt(PostAlgorithm):
    """Post-processing method by Hardt et al."""

    def __init__(self, unfavorable_label: int = 0, favorable_label: int = 1):
        """Init for Hardt."""
        super().__init__()
        self._unfavorable_label = unfavorable_label
        self._favorable_label = favorable_label
        self._random = RandomState(seed=888)

    @implements(PostAlgorithm)
    def run(
        self,
        train_predictions: pd.DataFrame,
        train: DataTuple,
        test_predictions: pd.DataFrame,
        test: TestTuple,
    ) -> pd.DataFrame:
        model_params = self._fit(train_predictions, train)
        return self._predict(model_params, test_predictions, test)

    def _fit(self, train_predictions: pd.DataFrame, train: DataTuple) -> OptimizeResult:
        # compute basic statistics
        fraction_s0 = (train.s[train.s.columns[0]].to_numpy() == 0).mean()
        fraction_s1 = 1 - fraction_s0

        s_col = train.s.columns[0]
        tprs = metric_per_sensitive_attribute(train_predictions, train, TPR())
        tpr0 = tprs[f"{s_col}_0"]
        tpr1 = tprs[f"{s_col}_1"]
        fnr0 = 1 - tpr0
        fnr1 = 1 - tpr1
        tnrs = metric_per_sensitive_attribute(train_predictions, train, TNR())
        tnr0 = tnrs[f"{s_col}_0"]
        tnr1 = tnrs[f"{s_col}_1"]
        fpr0 = 1 - tnr0
        fpr1 = 1 - tnr1

        # linear program has 4 decision variables:
        # [P[label_tilde = 1 | label_hat = 1, sensitive_attribute = 0];
        #  P[label_tilde = 1 | label_hat = 0, sensitive_attribute = 0];
        #  P[label_tilde = 1 | label_hat = 1, sensitive_attribute = 1];
        #  P[label_tilde = 1 | label_hat = 0, sensitive_attribute = 1]]
        # Coefficients of the linear objective function to be minimized.
        coeffs = np.array([fpr0 - tpr0, tnr0 - fnr0, fpr1 - tpr1, tnr1 - fnr1])

        # inequalilty_constraint_matrix: 2-D array which, when matrix-multiplied by x, gives the
        # values of the upper-bound inequality constraints at x
        # b_ub: 1-D array of values representing the upper-bound of each
        # inequality constraint (row) in A_ub.
        # Just to keep these between zero and one
        inequalilty_constraint_matrix: np.ndarray[np.float64] = np.array(
            [
                [1.0, 0.0, 0.0, 0.0],
                [-1.0, 0.0, 0.0, 0.0],
                [0.0, 1.0, 0.0, 0.0],
                [0.0, -1.0, 0.0, 0.0],
                [0.0, 0.0, 1.0, 0.0],
                [0.0, 0.0, -1.0, 0.0],
                [0.0, 0.0, 0.0, 1.0],
                [0.0, 0.0, 0.0, -1.0],
            ],
            dtype=np.float64,
        )
        b_ub: np.ndarray[np.float64] = np.array(
            [1.0, 0.0, 1.0, 0.0, 1.0, 0.0, 1.0, 0.0], dtype=np.float64
        )

        # Create boolean conditioning vectors for protected groups
        mask_s1 = train.s[train.s.columns[0]].to_numpy() == 1
        mask_s0 = train.s[train.s.columns[0]].to_numpy() == 0

        train_preds_numpy: np.ndarray = train_predictions[train_predictions.columns[0]].to_numpy()

        sconst = np.ravel(train_preds_numpy[mask_s1] == self._favorable_label)
        sflip = np.ravel(train_preds_numpy[mask_s1] == self._unfavorable_label)
        oconst = np.ravel(train_preds_numpy[mask_s0] == self._favorable_label)
        oflip = np.ravel(train_preds_numpy[mask_s0] == self._unfavorable_label)

        y_true = train.y.to_numpy().ravel()

        sm_tn = np.logical_and(sflip, y_true[mask_s1] == self._unfavorable_label)
        sm_fn = np.logical_and(sflip, y_true[mask_s1] == self._favorable_label)
        sm_fp = np.logical_and(sconst, y_true[mask_s1] == self._unfavorable_label)
        sm_tp = np.logical_and(sconst, y_true[mask_s1] == self._favorable_label)

        om_tn = np.logical_and(oflip, y_true[mask_s0] == self._unfavorable_label)
        om_fn = np.logical_and(oflip, y_true[mask_s0] == self._favorable_label)
        om_fp = np.logical_and(oconst, y_true[mask_s0] == self._unfavorable_label)
        om_tp = np.logical_and(oconst, y_true[mask_s0] == self._favorable_label)

        # A_eq - 2-D array which, when matrix-multiplied by x,
        # gives the values of the equality constraints at x.
        # b_eq - 1-D array of values representing the RHS of each equality constraint (row) in A_eq.
        # Used to impose equality of odds constraint
        a_eq = [
            [
                (np.mean(sconst * sm_tp) - np.mean(sflip * sm_tp)) / fraction_s1,
                (np.mean(sflip * sm_fn) - np.mean(sconst * sm_fn)) / fraction_s1,
                (np.mean(oflip * om_tp) - np.mean(oconst * om_tp)) / fraction_s0,
                (np.mean(oconst * om_fn) - np.mean(oflip * om_fn)) / fraction_s0,
            ],
            [
                (np.mean(sconst * sm_fp) - np.mean(sflip * sm_fp)) / (1 - fraction_s1),
                (np.mean(sflip * sm_tn) - np.mean(sconst * sm_tn)) / (1 - fraction_s1),
                (np.mean(oflip * om_fp) - np.mean(oconst * om_fp)) / (1 - fraction_s0),
                (np.mean(oconst * om_tn) - np.mean(oflip * om_tn)) / (1 - fraction_s0),
            ],
        ]

        b_eq = [
            (np.mean(oflip * om_tp) + np.mean(oconst * om_fn)) / fraction_s0
            - (np.mean(sflip * sm_tp) + np.mean(sconst * sm_fn)) / fraction_s1,
            (np.mean(oflip * om_fp) + np.mean(oconst * om_tn)) / (1 - fraction_s0)
            - (np.mean(sflip * sm_fp) + np.mean(sconst * sm_tn)) / (1 - fraction_s1),
        ]

        # Linear program
        return linprog(coeffs, A_ub=inequalilty_constraint_matrix, b_ub=b_ub, A_eq=a_eq, b_eq=b_eq)

    def _predict(
        self, model_params: OptimizeResult, test_predictions: pd.DataFrame, test: TestTuple
    ) -> pd.DataFrame:
        sp2p, sn2p, op2p, on2p = model_params.x

        # Create boolean conditioning vectors for protected groups
        mask_s1 = test.s[test.s.columns[0]].to_numpy() == 1
        mask_s0 = test.s[test.s.columns[0]].to_numpy() == 0

        test_preds_numpy: np.ndarray = test_predictions[test_predictions.columns[0]].to_numpy()

        # Randomly flip labels according to the probabilities in model_params
        self_fair_pred = test_preds_numpy[test.s[test.s.columns[0]].to_numpy() == 0].copy()
        self_pp_indices = (test_preds_numpy[mask_s1] == self._favorable_label).nonzero()[0]
        self_pn_indices = (test_preds_numpy[mask_s1] == self._unfavorable_label).nonzero()[0]
        self._random.shuffle(self_pp_indices)
        self._random.shuffle(self_pn_indices)

        n2p_indices = self_pn_indices[: int(len(self_pn_indices) * sn2p)]
        self_fair_pred[n2p_indices] = self._favorable_label
        p2n_indices = self_pp_indices[: int(len(self_pp_indices) * (1 - sp2p))]
        self_fair_pred[p2n_indices] = self._unfavorable_label

        othr_fair_pred = test_preds_numpy[mask_s0].copy()
        othr_pp_indices = (test_preds_numpy[mask_s0] == self._favorable_label).nonzero()[0]
        othr_pn_indices = (test_preds_numpy[mask_s0] == self._unfavorable_label).nonzero()[0]
        self._random.shuffle(othr_pp_indices)
        self._random.shuffle(othr_pn_indices)

        n2p_indices = othr_pn_indices[: int(len(othr_pn_indices) * on2p)]
        othr_fair_pred[n2p_indices] = self._favorable_label
        p2n_indices = othr_pp_indices[: int(len(othr_pp_indices) * (1 - op2p))]
        othr_fair_pred[p2n_indices] = self._unfavorable_label

        new_labels = np.zeros_like(test_preds_numpy, dtype=np.float64)
        new_labels[mask_s1] = self_fair_pred
        new_labels[mask_s0] = othr_fair_pred

        return pd.DataFrame(new_labels, columns=["preds"])

    @property
    def name(self) -> str:
        """Getter for algorithm name."""
        return "Hardt"



"""Abstract Base Class of all algorithms in the framework."""

from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Tuple, List
from abc import abstractmethod

from ethicml.common import implements
from ethicml.algorithms.algorithm_base import Algorithm, AlgorithmAsync, run_blocking
from ethicml.utility.data_structures import (
    DataTuple,
    TestTuple,
    PathTuple,
    TestPathTuple,
    write_as_feather,
)


class PreAlgorithm(Algorithm):
    """Abstract Base Class for all algorithms that do pre-processing."""

    @abstractmethod
    def run(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        """Generate fair features with the given data.

        Args:
            train: training data
            test: test data
        """

    def run_test(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        """Run with reduced training set so that it finishes quicker."""
        train_testing = train.get_subset()
        return self.run(train_testing, test)


class PreAlgorithmAsync(PreAlgorithm, AlgorithmAsync):
    """Pre-Algorithm that can be run blocking and asynchronously."""

    @implements(PreAlgorithm)
    def run(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        """Generate fair features with the given data by running as a blocking function.

        Args:
            train: training data
            test: test data
        """
        return run_blocking(self.run_async(train, test))

    async def run_async(self, train: DataTuple, test: TestTuple) -> Tuple[DataTuple, TestTuple]:
        """Generate fair features with the given data asynchronously.

        Args:
            train: training data
            test: test data
        """
        with TemporaryDirectory() as tmpdir:
            tmp_path = Path(tmpdir)
            # write data to files
            train_paths, test_paths = write_as_feather(train, test, tmp_path)

            # ========================== generate commandline arguments ===========================
            transformed_train_paths = PathTuple(
                x=tmp_path / "transform_train_x.feather",
                s=tmp_path / "transform_train_s.feather",
                y=tmp_path / "transform_train_y.feather",
                name=train.name if train.name is not None else "",
            )
            transformed_test_paths = TestPathTuple(
                x=tmp_path / "transform_test_x.feather",
                s=tmp_path / "transform_test_s.feather",
                name=test.name if test.name is not None else "",
            )
            cmd = self._script_command(
                train_paths, test_paths, transformed_train_paths, transformed_test_paths
            )

            # ============================= run the generated command =============================
            await self._call_script(cmd)

            # ================================== load results =====================================
            transformed_train = transformed_train_paths.load_from_feather()
            transformed_test = transformed_test_paths.load_from_feather()

        # prefix the name of the algorithm to the dataset name
        transformed_train = transformed_train.replace(
            name=None if train.name is None else f"{self.name}: {train.name}"
        )
        transformed_test = transformed_test.replace(
            name=None if test.name is None else f"{self.name}: {test.name}"
        )
        return transformed_train, transformed_test

    @abstractmethod
    def _script_command(
        self,
        train_paths: PathTuple,
        test_paths: TestPathTuple,
        new_train_paths: PathTuple,
        new_test_paths: TestPathTuple,
    ) -> List[str]:
        """The command that will run the script."""
